﻿/*Copyright 2015 Unknown Developer

Code Found by Wojciech Christopher Peter Fiałkiewicz in second half of 2015 somewhere in the internet.
 */

using System;
using System.Drawing;
using System.Drawing.Imaging;
using System.Runtime.InteropServices;

namespace AviFile
{
    public class VideoStream : AviStream
	{
		private int frameSize;
        public int FrameSize {
            get { return frameSize; }
        }

        protected double frameRate;
        public double FrameRate {
            get{ return frameRate; }
		}

		private int width;
		public int Width{
			get{ return width; }
		}

		private int height;
		public int Height{
			get{ return height; }
		}

		private Int16 countBitsPerPixel;
        public Int16 CountBitsPerPixel {
            get{ return countBitsPerPixel; }
		}

		/// <summary>count of frames in the stream</summary>
		protected int countFrames = 0;
		public int CountFrames{
			get{ return countFrames; }
		}

        /// <summary>Palette for indexed frames</summary>
        protected Avi.RGBQUAD[] palette;
        public Avi.RGBQUAD[] Palette{
            get { return palette; }
        }

		/// <summary>initial frame index</summary>
		/// <remarks>Added by M. Covington</remarks>
		protected int firstFrame = 0;
		public int FirstFrame
		{
			get { return firstFrame; }
		}

        private Avi.AVICOMPRESSOPTIONS compressOptions;
        public Avi.AVICOMPRESSOPTIONS CompressOptions {
            get { return compressOptions; }
        }
        
        public VideoStream(int aviFile, bool writeCompressed, double frameRate, Bitmap firstFrame) {
            Initialize(aviFile, writeCompressed, frameRate, firstFrame);
            CreateStream();
			AddFrame(firstFrame);
		}

        /// <summary>Copy a palette</summary>
        /// <param name="template">Original palette</param>
        
        
        private void CopyPalette(ColorPalette template)
        {
            this.palette = new Avi.RGBQUAD[template.Entries.Length];

            for (int n = 0; n < this.palette.Length; n++)
            {
                if (n < template.Entries.Length)
                {
                    this.palette[n].rgbRed = template.Entries[n].R;
                    this.palette[n].rgbGreen = template.Entries[n].G;
                    this.palette[n].rgbBlue = template.Entries[n].B;
                }
                else
                {
                    this.palette[n].rgbRed = 0;
                    this.palette[n].rgbGreen = 0;
                    this.palette[n].rgbBlue = 0;
                }
            }
        }
          
        private void Initialize(int aviFile, bool writeCompressed, double frameRate, Bitmap firstFrameBitmap) {
            this.aviFile = aviFile;
            this.writeCompressed = writeCompressed;
            this.frameRate = frameRate;
			this.firstFrame = 0;

            CopyPalette(firstFrameBitmap.Palette);

			BitmapData bmpData = firstFrameBitmap.LockBits(new Rectangle(
				0, 0, firstFrameBitmap.Width, firstFrameBitmap.Height),
				ImageLockMode.ReadOnly, firstFrameBitmap.PixelFormat);

            this.frameSize = bmpData.Stride * bmpData.Height;
			this.width = firstFrameBitmap.Width;
			this.height = firstFrameBitmap.Height;
			this.countBitsPerPixel = ConvertPixelFormatToBitCount(firstFrameBitmap.PixelFormat);

			firstFrameBitmap.UnlockBits(bmpData);
        }

        private Int16 ConvertPixelFormatToBitCount(PixelFormat format)
        {
			String formatName = format.ToString();
			if(formatName.Substring(0, 6) != "Format"){
				throw new Exception("Unknown pixel format: "+formatName);
			}

			formatName = formatName.Substring(6, 2);
			Int16 bitCount = 0;
			if( Char.IsNumber(formatName[1]) ){	//16, 32, 48
				bitCount = Int16.Parse(formatName);
			}else{								//4, 8
				bitCount = Int16.Parse(formatName[0].ToString());
			}

			return bitCount;
		}
        
        private void GetRateAndScale(ref double frameRate, ref int scale)
        {
            scale = 1;
            while (frameRate != (long)frameRate) {
                frameRate = frameRate * 10;
                scale *= 10;
            }
        }

        private void CreateStreamWithoutFormat()
        {
            int scale = 1;
            double rate = frameRate;
            GetRateAndScale(ref rate, ref scale);

            Avi.AVISTREAMINFO strhdr = new Avi.AVISTREAMINFO();
            strhdr.fccType = Avi.mmioStringToFOURCC("vids", 0);
            strhdr.fccHandler = Avi.mmioStringToFOURCC("CVID", 0);
            strhdr.dwFlags = 0;
            strhdr.dwCaps = 0;
            strhdr.wPriority = 0;
            strhdr.wLanguage = 0;
            strhdr.dwScale = (int)scale;
            strhdr.dwRate = (int)rate; // Frames per Second
            strhdr.dwStart = 0;
            strhdr.dwLength = 0;
            strhdr.dwInitialFrames = 0;
            strhdr.dwSuggestedBufferSize = frameSize; //height_ * stride_;
            strhdr.dwQuality = -1;        //default
            strhdr.dwSampleSize = 0;
            strhdr.rcFrame.top = 0;
            strhdr.rcFrame.left = 0;
            strhdr.rcFrame.bottom = (uint)height;
            strhdr.rcFrame.right = (uint)width;
            strhdr.dwEditCount = 0;
            strhdr.dwFormatChangeCount = 0;
            strhdr.szName = new UInt16[64];

            int result = Avi.AVIFileCreateStream(aviFile, out aviStream, ref strhdr);

            if (result != 0) {
                throw new Exception("Exception in AVIFileCreateStream: " + result.ToString());
            }
        }

                /// <summary>Create a new stream</summary>
        
        private void CreateStream()
        {
            CreateStreamWithoutFormat();

            if (writeCompressed) {
                CreateCompressedStream();
            } else {
                //SetFormat(aviStream, 0);
            }
        }

	    private void CreateCompressedStream()
        {
			//display the compression options dialog...
			Avi.AVICOMPRESSOPTIONS_CLASS options = new Avi.AVICOMPRESSOPTIONS_CLASS();
			options.fccType = (uint)Avi.streamtypeVIDEO;

            options.lpParms = IntPtr.Zero;
			options.lpFormat = IntPtr.Zero;
			Avi.AVISaveOptions(IntPtr.Zero, Avi.ICMF_CHOOSE_KEYFRAME | Avi.ICMF_CHOOSE_DATARATE, 1, ref aviStream, ref options);
			
           //get the compressed stream
            this.compressOptions = options.ToStruct();
            int result = Avi.AVIMakeCompressedStream(out compressedStream, aviStream, ref compressOptions, 0);
            if(result != 0) {
				throw new Exception("Exception in AVIMakeCompressedStream: "+result.ToString());
			}

            Avi.AVISaveOptionsFree(1, ref options);
            SetFormat(compressedStream, 0);
		}

        public void AddFrame(Bitmap bmp)
        {
            bmp.RotateFlip(RotateFlipType.RotateNoneFlipY);

            // NEW 2012-11-10
            if (countFrames == 0)
            {
                CopyPalette(bmp.Palette);
                SetFormat(writeCompressed ? compressedStream : StreamPointer, countFrames);
            }

			BitmapData bmpDat = bmp.LockBits(
				new Rectangle(
				0,0, bmp.Width, bmp.Height),
				ImageLockMode.ReadOnly, bmp.PixelFormat);

            int result = Avi.AVIStreamWrite(writeCompressed ? compressedStream : StreamPointer,
                countFrames, 1, 
				bmpDat.Scan0, 
				(Int32)(bmpDat.Stride * bmpDat.Height), 
				0, 0, 0);

			if (result!= 0) {
				throw new Exception("Exception in VideoStreamWrite: "+result.ToString());
			}

			bmp.UnlockBits(bmpDat);

			countFrames++;
		}
	   
        private void SetFormat(IntPtr aviStream, int writePosition)
        {
	        Avi.BITMAPINFO bi = new Avi.BITMAPINFO();
            bi.bmiHeader.biWidth = width;
            bi.bmiHeader.biHeight = height;
            bi.bmiHeader.biPlanes = 1;
            bi.bmiHeader.biBitCount = countBitsPerPixel;
            bi.bmiHeader.biSizeImage = frameSize;
            bi.bmiHeader.biSize = Marshal.SizeOf(bi.bmiHeader);

            if (countBitsPerPixel < 24)
            {
                bi.bmiHeader.biClrUsed = this.palette.Length;
                bi.bmiHeader.biClrImportant = this.palette.Length;
                bi.bmiColors = new Avi.RGBQUAD[this.palette.Length];
                this.palette.CopyTo(bi.bmiColors, 0);
                bi.bmiHeader.biSize += bi.bmiColors.Length * Avi.RGBQUAD_SIZE;
            }

            int result = Avi.AVIStreamSetFormat(aviStream, writePosition, ref bi, bi.bmiHeader.biSize);
            if(result != 0){ throw new Exception("Error in VideoStreamSetFormat: "+ result.ToString("X")); }
		}
        	
		public override void ExportStream(String fileName){
			Avi.AVICOMPRESSOPTIONS_CLASS opts = new Avi.AVICOMPRESSOPTIONS_CLASS();
			opts.fccType = (uint)Avi.streamtypeVIDEO;
			opts.lpParms = IntPtr.Zero;
			opts.lpFormat = IntPtr.Zero;
            IntPtr streamPointer = StreamPointer;
            Avi.AVISaveOptions(IntPtr.Zero, Avi.ICMF_CHOOSE_KEYFRAME | Avi.ICMF_CHOOSE_DATARATE, 1, ref streamPointer, ref opts);
            Avi.AVISaveOptionsFree(1, ref opts);

			Avi.AVISaveV(fileName, 0, 0, 1, ref aviStream, ref opts);
		}
    }
}
